"""Tests for class :class:`ncdata.NcVariable`.

There is almost no actual behaviour, but we can test some constructor behaviours,
such as valid calling options and possibilities for data and dtype.
"""
import re

import dask.array as da
import numpy as np
import pytest

from ncdata import NcAttribute, NcData, NcVariable


class Test_NcVariable__init__:
    def test_minimum_args(self):
        # We can create a variable with no args.
        name = "varname"
        sample = NcVariable("varname")
        # No data, no dtype.  Variables don't have 'shape' anyway
        assert sample.name is name
        assert sample.dimensions == ()
        assert sample.data is None
        assert sample.dtype is None
        assert sample.attributes == {}

    def test_never_nameless(self):
        # Can't create a variable without a name.
        msg = "required positional argument: 'name'"
        with pytest.raises(TypeError, match=msg):
            _ = NcVariable()

    def test_allargs(self):
        # Since there is no type checking, you can put whatever you like in the "slots".
        name = "data_name"
        dims = {f"dimname_{i}": f"dim_{i}" for i in range(3)}
        vars = {f"varname_{i}": f"var_{i}" for i in range(5)}
        attrs = {f"varname_{i}": f"var_{i}" for i in range(6)}
        grps = {f"groupname_{i}": f"group_{i}" for i in range(3)}
        sample1 = NcData(
            name=name,
            dimensions=dims,
            variables=vars,
            attributes=attrs,
            groups=grps,
        )
        # Nothing is copied : all contents are simply references to the inputs
        assert sample1.name is name
        assert sample1.dimensions is dims
        assert sample1.variables is vars
        assert sample1.groups is grps
        assert sample1.attributes is attrs
        # Also check construction with arguments alone (no keywords).
        sample2 = NcData(name, dims, vars, attrs, grps)
        # result is new object, but all contents are references identical to the other
        assert sample2 is not sample1
        for name in [n for n in dir(sample1) if n[0] != "_"]:
            assert getattr(sample2, name) is getattr(sample1, name)

    @pytest.mark.parametrize("with_data", [False, True])
    def test_dtype__with_and_without_data(self, with_data):
        test_dtype_nonfunctional = "invalid_dtype"
        data = np.arange(2.0) if with_data else None
        var = NcVariable("x", data=data, dtype=test_dtype_nonfunctional)
        expect_dtype = data.dtype if with_data else test_dtype_nonfunctional
        assert var.dtype == expect_dtype

    @pytest.mark.parametrize(
        # All of these cases should deliver data wrapped as np.asanyarray()
        "data",
        [
            12,
            3.4,
            "a",
            "bcd",
            [1, 2],
            [(3.4, 4), (5, 6)],
            ["xxx", "yyy", "zzz"],
        ],
    )
    def test_data__pythontypes(self, data):
        var = NcVariable("x", data=data, dtype="ineffective")
        expect_array = np.asanyarray(data)
        assert isinstance(var.data, np.ndarray)
        assert var.dtype == expect_array.dtype
        assert np.all(var.data == expect_array)
        if isinstance(data, np.ndarray):
            # Actual array data is stored directly
            assert var.data is data

    def test_data__listsofstrings(self):
        # handling of a ragged string-length is a specific case worth exploring
        data = [["one", "two", "three"], ["4", "5", "6"]]
        var = NcVariable("x", data=data)
        assert isinstance(var.data, np.ndarray)
        assert var.data.shape == (2, 3)
        assert var.dtype == var.data.dtype == np.dtype("U5")

    @pytest.mark.parametrize(
        "dtype", ["i4", "i8", "f4", "f8", "u1", "i1", "U4"]
    )
    def test_data__numpy(self, dtype):
        elem = "abc" if dtype.startswith("U") else 12
        dtype = np.dtype(dtype)
        arr = np.zeros((2, 3), dtype=dtype)
        arr[:] = elem
        # The variable should preserve the original of a contained numpy array, and adopt
        # it's dtype as it's own
        var = NcVariable("x", data=arr)
        assert var.data is arr  # not copied or re-formed
        assert var.dtype == dtype

    @pytest.mark.parametrize("dtype", ["i4", "i8", "f4", "f8", "u1", "i1"])
    def test_data__numpy_scalar(self, dtype):
        # As above, but for scalars.
        dtype = np.dtype(dtype)
        value = np.array(1.0, dtype=dtype)
        var = NcVariable("x", data=value)
        assert var.data is value
        assert var.dtype == dtype

    def test_data__dask(self):
        # Check that lazy data is also contained directly, without adjustment or
        # modification
        arr = da.ones((2, 3))
        var = NcVariable("x", data=arr)
        assert var.data is arr


class TestNcVariable__str__repr:
    def test_nameonly(self):
        var = NcVariable("varname")
        result = str(var)
        expected = "<NcVariable(<no-dtype>): varname()>"
        assert result == expected

    def test_dimsnoargs(self):
        var = NcVariable("var_w_dims", ["dim0", "dim1"])
        result = str(var)
        expected = "<NcVariable(<no-dtype>): var_w_dims(dim0, dim1)>"
        assert result == expected

    def test_oneattr(self):
        var = NcVariable(
            "var_w_attrs", attributes={"a1": NcAttribute("a1", 1)}
        )
        result = str(var)
        expected = "\n".join(
            [
                "<NcVariable(<no-dtype>): var_w_attrs()",
                "    var_w_attrs:a1 = 1",
                ">",
            ]
        )
        assert result == expected

    def test_multiattrs(self):
        var = NcVariable(
            "var_multi",
            attributes={
                "a1": NcAttribute("a1", 1),
                "a2": NcAttribute("a2", ["one", "three"]),
            },
        )
        result = str(var)
        expected = "\n".join(
            [
                "<NcVariable(<no-dtype>): var_multi()",
                "    var_multi:a1 = 1",
                "    var_multi:a2 = ['one', 'three']",
                ">",
            ]
        )
        assert result == expected

    def test_str(self):
        var = NcVariable(
            "var",
            dimensions=("x", "y"),
            data=np.ones((2, 3), dtype=np.float32),
        )
        result = str(var)
        expected = "<NcVariable(float32): var(x, y)>"
        assert result == expected

    def test_repr(self):
        var = NcVariable(
            "var",
            dimensions=("x", "y"),
            attributes={"a1": NcAttribute("a1", 1)},
        )
        result = repr(var)
        expected = f"<ncdata._core.NcVariable object at 0x{id(var):012x}>"
        assert result == expected
